/*

  Creaturepod

  Author: Joe Grand [www.grandideastudio.com]

  Program Description:

  My kids wanted a real life Creaturepod from Wild Kratts. It didn't exist.
  We made one.

  Arduino Pro Mini 328 3.3V/8MHz -> Adafruit 2.2" TFT Breakout w/ SD card -> 
  DRA818U UHF Band Voice Transceiver Module
  
  Revisions:

  1.0 (September 22, 2015): Initial release
  
*/

/***************************************************
  This library works with the Adafruit 2.2" TFT Breakout w/SD card
  ----> http://www.adafruit.com/products/1480
 
  Check out the links above for our tutorials and wiring diagrams
  These displays use SPI to communicate, 4 or 5 pins are required to
  interface (RST is optional)
  Adafruit invests time and resources providing this open source code,
  please support Adafruit and open-source hardware by purchasing
  products from Adafruit!

  Written by Limor Fried/Ladyada for Adafruit Industries.
  MIT license, all text above must be included in any redistribution
 ****************************************************/

#include <SPI.h>                  // SPI (Serial Peripheral Interface) library
#include <SD.h>                   // SD (SecureDigital) library
#include <SoftwareSerial.h>       // SoftwareSerial library
#include <Adafruit_GFX.h>         // Core graphics library
#include <Adafruit_ILI9340.h>     // Hardware-specific library

// Pin definitions
#define TFT_RST     8   // LCD
#define TFT_DC      7
#define TFT_CS      10

#define SD_CS       9   // microSD

#define SW_L        4   // Channel select switches
#define SW_C        3
#define SW_R        2

#define RADIO_RX    5  // Serial input (connects to the DRA818U's TXD pin)
#define RADIO_TX    6  // Serial output (connects to the DRA818U's RXD pin)


Adafruit_ILI9340 tft = Adafruit_ILI9340(TFT_CS, TFT_DC, TFT_RST);   // Set up TFT LCD interface
SoftwareSerial draSerial =  SoftwareSerial(RADIO_RX, RADIO_TX);     // Set up serial port object for communication to DRA818U

// user configuration
// only one user can be defined at a time
#define __USER      U_ALICE
  
#define U_ALICE     1
#define U_BOB       2
#define U_EVE   3
#define U_MALLORY   4

// frequency definitions
// ensure that the frequencies used are in the open spectrum designated 
// for unlicensed communications or that all users have the proper licenses
#define F_ALICE     433.0750
#define F_BOB       433.1000
#define F_EVE   433.1250
#define F_MALLORY   433.1500

// images are pre-loaded onto the microSD card
#define P_LOGO      "logo.bmp"
#define P_ALICE     "alice.bmp"
#define P_BOB       "bob.bmp"
#define P_EVE       "eve.bmp"
#define P_MALLORY   "mallory.bmp"

// set defaults
char* bitmap = P_LOGO; 

#if (__USER == U_ALICE)
  #define F_TX  F_ALICE
  #define F_RX  F_ALICE
#elif (__USER == U_BOB)
  #define F_TX  F_BOB
  #define F_RX  F_BOB
#elif (__USER == U_EVE)
  #define F_TX  F_EVE
  #define F_RX  F_EVE
#elif (__USER == U_MALLORY)
  #define F_TX  F_MALLORY
  #define F_RX  F_MALLORY
#endif

// set parameters for DRA818U
// Code based on http://www.kh-gps.de/dra.htm and modified for the UHF module
int bw = 1;                 // bandwidth in KHz ( 0 = 12.5 or 1 = 25 )
float ftx = F_TX;           // tx frequency in MHz (400.0000 - 470.0000)
float frx = F_RX;           // rx frequency in MHz (400.0000 - 470.0000) 
String tx_ctcss = "0000";   // ctcss frequency ( 0000 - 0038 ); 0000 = "no CTCSS"  
String rx_ctcss = "0000";   // ctcss frequency ( 0000 - 0038 ); 0000 = "no CTCSS"  
int squ = 2;                // squelch level  ( 0 - 8 ); 0 = "open"  


void setup(void) {
  // Initialize GPIO
  pinMode(SW_L, INPUT);
  pinMode(SW_C, INPUT);
  pinMode(SW_R, INPUT);
  pinMode(RADIO_RX, INPUT);
  pinMode(RADIO_TX, OUTPUT);

  Serial.begin(9600);
  while (!Serial);   // wait until ready
  Serial.println("\n\nActivating the Creaturepod!\n");

  Serial.print("Initializing SD card...");
  if (!SD.begin(SD_CS)) {
    Serial.println("Failed!");
    return;
  }
  Serial.println("OK!");

  tft.begin();
  bmpDraw(bitmap, 0, 0);   // display default bitmap

  Serial.print("Initializing DRA818U radio transceiver module...");
  draSerial.begin(9600);  // set baud rate for SoftwareSerial port
  delay(10);              // short delay
  draGroupSet();          // set radio defaults
  Serial.println("OK!");

  // Display transceiver settings
  Serial.print("Bandwidth: ");
    if (bw == 0) Serial.println("12.5 kHz");
    else         Serial.println("25.0 kHz");
  Serial.print("TX frequency: "); Serial.print(ftx, 4); Serial.println(" MHz");
  Serial.print("TX CTCSS value: "); Serial.println(tx_ctcss);
  Serial.print("RX frequency: "); Serial.print(frx, 4); Serial.println(" MHz");
  Serial.print("RX CTCSS value: "); Serial.println(rx_ctcss);
  Serial.print("Squelch level: "); Serial.println(squ);
}

void loop() {
  if (!digitalRead(SW_L)) // check for left button press
  {
    delay(100);                // poor man's debounce
    while(!digitalRead(SW_L)); // wait for button to be released

    // set frequency and display picture of selected user
    #if (__USER == U_ALICE || __USER == U_BOB)      
      if (ftx == F_EVE) { ftx = F_TX; bitmap = P_LOGO; }    // if user was already selected, go back to main screen
      else              { ftx = F_EVE; bitmap = P_EVE; }   // set new parameters 
    #elif (__USER == U_EVE || U_MALLORY)
      if (ftx == F_ALICE) { ftx = F_TX; bitmap = P_LOGO; } 
      else                { ftx = F_ALICE; bitmap = P_ALICE; }
    #endif

   Serial.print("Setting TX frequency to "); Serial.print(ftx, 4); Serial.print(" MHz...");
   draGroupSet();
   Serial.println("OK!");
   bmpDraw(bitmap, 0, 0);
  }
  else if (!digitalRead(SW_C)) // check for center button press
  {
    delay(100);                // poor man's debounce
    while(!digitalRead(SW_C)); // wait for button to be released
    
    // set frequency and display picture of selected user
    #if (__USER == U_ALICE || __USER == U_EVE || __USER == U_MALLORY)      
      if (ftx == F_BOB) { ftx = F_TX; bitmap = P_LOGO; }   // if user was already selected, go back to main screen
      else              { ftx = F_BOB; bitmap = P_BOB; }       // set new parameters 
    #elif (__USER == U_BOB)
      if (ftx == F_ALICE) { ftx = F_TX; bitmap = P_LOGO; } 
      else                { ftx = F_ALICE; bitmap = P_ALICE; }  
    #endif
   
   Serial.print("Setting TX frequency to "); Serial.print(ftx, 4); Serial.print(" MHz...");
   draGroupSet();
   Serial.println("OK!");
   bmpDraw(bitmap, 0, 0);  
  }
  else if (!digitalRead(SW_R)) // check for right button press
  {
    delay(100);                // poor man's debounce    
    while(!digitalRead(SW_R)); // wait for button to be released

    // set frequency and display picture of selected user
    #if (__USER == U_ALICE || __USER == U_BOB || __USER == U_EVE)  
      if (ftx == F_MALLORY) { ftx = F_TX; bitmap = P_LOGO; }   // if user was already selected, go back to main screen
      else                  { ftx = F_MALLORY; bitmap = P_MALLORY; }   // set new parameters 
    #elif (__USER == U_MALLORY)
      if (ftx == F_EVE) { ftx = F_TX; bitmap = P_LOGO; }     
      else              { ftx = F_EVE; bitmap = P_EVE; } 
    #endif 
   
   Serial.print("Setting TX frequency to "); Serial.print(ftx, 4); Serial.print(" MHz...");
   draGroupSet();
   Serial.println("OK!");
   bmpDraw(bitmap, 0, 0);
  }
}

/***************************************************/

void draGroupSet()   // DRA818U group setting command
{
  draSerial.print("AT+DMOSETGROUP="); // begin message 
  draSerial.print(bw, 1);
  draSerial.print(","); 
  draSerial.print(ftx, 4); 
  draSerial.print(","); 
  draSerial.print(frx, 4); 
  draSerial.print(","); 
  draSerial.print(tx_ctcss); 
  draSerial.print(","); 
  draSerial.print(squ); 
  draSerial.print(","); 
  draSerial.println(rx_ctcss);
  while (draSerial.read() != '0');    // wait for response from module indicating success
}

/***************************************************/

// This function opens a Windows Bitmap (BMP) file and
// displays it at the given coordinates.  It's sped up
// by reading many pixels worth of data at a time
// (rather than pixel by pixel).  Increasing the buffer
// size takes more of the Arduino's precious RAM but
// makes loading a little faster.  20 pixels seems a
// good balance.

#define BUFFPIXEL 20

void bmpDraw(char *filename, uint16_t x, uint16_t y) {

  File     bmpFile;
  int      bmpWidth, bmpHeight;   // W+H in pixels
  uint8_t  bmpDepth;              // Bit depth (currently must be 24)
  uint32_t bmpImageoffset;        // Start of image data in file
  uint32_t rowSize;               // Not always = bmpWidth; may have padding
  uint8_t  sdbuffer[3*BUFFPIXEL]; // pixel buffer (R+G+B per pixel)
  uint8_t  buffidx = sizeof(sdbuffer); // Current position in sdbuffer
  boolean  goodBmp = false;       // Set to true on valid header parse
  boolean  flip    = true;        // BMP is stored bottom-to-top (true)
  int      w, h, row, col;
  uint8_t  r, g, b;
  uint32_t pos = 0/*, startTime = millis()*/;

  if((x >= tft.width()) || (y >= tft.height())) return;

  //Serial.println();
  Serial.print("Loading image '");
  Serial.print(filename);
  Serial.print("\'...");

  // Open requested file on SD card
  if ((bmpFile = SD.open(filename)) == NULL) {
    Serial.print("File not found");
    return;
  }

  // Parse BMP header
  if(read16(bmpFile) == 0x4D42) { // BMP signature
    //Serial.print("File size: "); Serial.println(read32(bmpFile));
    read32(bmpFile);
    (void)read32(bmpFile); // Read & ignore creator bytes
    bmpImageoffset = read32(bmpFile); // Start of image data
    //Serial.print("Image Offset: "); Serial.println(bmpImageoffset, DEC);
    // Read DIB header
    //Serial.print("Header size: "); Serial.println(read32(bmpFile));
    read32(bmpFile);
    bmpWidth  = read32(bmpFile);
    bmpHeight = read32(bmpFile);
    if(read16(bmpFile) == 1) { // # planes -- must be '1'
      bmpDepth = read16(bmpFile); // bits per pixel
      //Serial.print("Bit Depth: "); Serial.println(bmpDepth);
      if((bmpDepth == 24) && (read32(bmpFile) == 0)) { // 0 = uncompressed

        goodBmp = true; // Supported BMP format -- proceed!
        /*Serial.print("Image size: ");
        Serial.print(bmpWidth);
        Serial.print('x');
        Serial.println(bmpHeight);*/

        // BMP rows are padded (if needed) to 4-byte boundary
        rowSize = (bmpWidth * 3 + 3) & ~3;

        // If bmpHeight is negative, image is in top-down order.
        // This is not canon but has been observed in the wild.
        if(bmpHeight < 0) {
          bmpHeight = -bmpHeight;
          flip      = false;
        }

        // Crop area to be loaded
        w = bmpWidth;
        h = bmpHeight;
        if((x+w-1) >= tft.width())  w = tft.width()  - x;
        if((y+h-1) >= tft.height()) h = tft.height() - y;

        // Set TFT address window to clipped image bounds
        tft.setAddrWindow(x, y, x+w-1, y+h-1);

        for (row=0; row<h; row++) { // For each scanline...

          // Seek to start of scan line.  It might seem labor-
          // intensive to be doing this on every line, but this
          // method covers a lot of gritty details like cropping
          // and scanline padding.  Also, the seek only takes
          // place if the file position actually needs to change
          // (avoids a lot of cluster math in SD library).
          if(flip) // Bitmap is stored bottom-to-top order (normal BMP)
            pos = bmpImageoffset + (bmpHeight - 1 - row) * rowSize;
          else     // Bitmap is stored top-to-bottom
            pos = bmpImageoffset + row * rowSize;
          if(bmpFile.position() != pos) { // Need seek?
            bmpFile.seek(pos);
            buffidx = sizeof(sdbuffer); // Force buffer reload
          }

          for (col=0; col<w; col++) { // For each pixel...
            // Time to read more pixel data?
            if (buffidx >= sizeof(sdbuffer)) { // Indeed
              bmpFile.read(sdbuffer, sizeof(sdbuffer));
              buffidx = 0; // Set index to beginning
            }

            // Convert pixel from BMP to TFT format, push to display
            b = sdbuffer[buffidx++];
            g = sdbuffer[buffidx++];
            r = sdbuffer[buffidx++];
            tft.pushColor(tft.Color565(r,g,b));
          } // end pixel
        } // end scanline
        Serial.println("OK!");
        /*Serial.print("Loaded in ");
        Serial.print(millis() - startTime);
        Serial.println(" ms");*/
      } // end goodBmp
    }
  }

  bmpFile.close();
  if(!goodBmp) Serial.println("BMP format not recognized.");
}

// These read 16- and 32-bit types from the SD card file.
// BMP data is stored little-endian, Arduino is little-endian too.
// May need to reverse subscript order if porting elsewhere.

uint16_t read16(File & f) {
  uint16_t result;
  ((uint8_t *)&result)[0] = f.read(); // LSB
  ((uint8_t *)&result)[1] = f.read(); // MSB
  return result;
}

uint32_t read32(File & f) {
  uint32_t result;
  ((uint8_t *)&result)[0] = f.read(); // LSB
  ((uint8_t *)&result)[1] = f.read();
  ((uint8_t *)&result)[2] = f.read();
  ((uint8_t *)&result)[3] = f.read(); // MSB
  return result;
}
